/////////////////////////////////////////////////////////////////////////////////////
// Filename : d_mem.h
// Project  : Data Memory Model
// Author   : GoodKook / goodkook@nms.anslab.co.kr
// Company  : AnsLab Co. / http://www.anslab.co.kr
// Revision :
// Comment  : Memory Model
//      1. Synchronous Read-Write (rising-edge of in_clk, out_clk)
//      2. Data Input/Output (Dual-Port)
//      3. Write enable control input (WEN='1' for write, WEN='0' for Read)
//      4. Memory Download/Dump utility
//      5. Parametized-Reusable (MEM_SIZE/MEM_WIDTH/MEM_ADDR)
//
#ifndef D_MEM_INC
#define D_MEM_INC

#include "systemc.h"
#include "stdlib.h"
#include "malloc.h"

#define MEM_SIZE    3584    // Number Words
#define MEM_ADDR    12      // Width of Address (MEM_SIZE=2^MEM_ADDR)
#define MEM_WIDTH   16      // Width of Data

#define DOWNLOAD_FILE   "download.dat"
#define DUMP_FILE       "dump.dat"

SC_MODULE(d_mem)
{
    sc_out<sc_bv<MEM_WIDTH> >   q;              // Data Output Bus
    // Memory control : Scale Factor Unit
    sc_in<sc_bit>               scl_inclk;
    sc_in<sc_bv<MEM_WIDTH> >    scl_data;       // Data Input Bus
    sc_in<sc_bit>               scl_ready;
    sc_in<sc_bit>               scl_wen;        // Write enable-Scale Factor
    sc_in<sc_bv<MEM_ADDR> >     scl_address;    // Address Bus-Scale Factor
    // Memory control : Huffman Decoder Unit
    sc_in<sc_bit>               huff_inclk;
    sc_in<sc_bv<MEM_WIDTH> >    huff_data;      // Data Input Bus
    sc_in<sc_bit>               huff_ready;
    sc_in<sc_bit>               huff_wen;       // Write enable-Huffman Decoder
    sc_in<sc_bv<MEM_ADDR> >     huff_address;   // Address Bus-Huffman Decoder

    sc_in<bool>                 download;       // Memory Download
    sc_in<bool>                 dump;           // Memory Dump

    // Processes ///////////////////////////////////////////////////////////
    void proc_d_mem_read();
    void proc_d_mem_write();

    void proc_d_mem_download();
    void proc_d_mem_dump();

    void proc_address_mux();
    void proc_wen_mux();

    char* d_mem::get_data(char *buff, char* vector);
    int d_mem::get_addr(char *buff);

    // Memory Module /////////////////////////////////////////////////////
    void *pMem;

    // Signals ///////////////////////////////////////////////////////////
    public:
    sc_signal<sc_bit>               in_clk;
    sc_signal<sc_bit>               wen;
    sc_signal<sc_bv<MEM_ADDR> >     address;
    sc_signal<sc_bv<MEM_WIDTH> >    data;
    sc_signal<sc_bv<MEM_WIDTH> >    t_q;

    // Constructor ///////////////////////////////////////////////////////
    SC_CTOR(d_mem)
    {
        // Processes /////////////////////////////////////////
        SC_METHOD(proc_wen_mux);
        sensitive
            << scl_inclk << scl_wen
            << huff_inclk << huff_wen;

        SC_METHOD(proc_address_mux);
        sensitive
            << scl_data << scl_address << scl_ready
            << huff_data << huff_address << huff_ready;

        SC_METHOD(proc_d_mem_write);
        sensitive_pos << in_clk;

        SC_METHOD(proc_d_mem_read);
        sensitive << address << wen;

        SC_METHOD(proc_d_mem_download);
        sensitive << download;

        SC_METHOD(proc_d_mem_dump);
        sensitive << dump;

        // Memory Allocate
        pMem = calloc(MEM_SIZE, sizeof(sc_uint<MEM_WIDTH>));
    }
};

#endif
