///////////////////////////////////////////////////////////////////////////
// Filename : Mp3Common.h
// Project  : Ans_MP3_Player/MP3_Decoder
// Author   : GoodKook / goodkook@nms.anslab.co.kr
// Company  : AnsLab Co. / http://www.anslab.co.kr
// Revision : 5-4-01: Modified for interface to untimed SystemC testbench
//                    - John Stickley, Ikos Systems, Inc. (johns@ikos.com)
// Comment  :
//

#ifndef MP3_COMMON_H
#define MP3_COMMON_H

#define TRACE_TIME_UNIT -9      // nS
#define CLOCK_PERIOD    20E-9   // 20 nS
#define RESET_CYCLES    8

#define PI  3.14159265358979

#define FRAME_SIZE  1024
#define BUFFER_SIZE 4096

#define MIN(A, B)   ((A) < (B) ? (A) : (B))
#define MAX(A, B)   ((A) > (B) ? (A) : (B))

// Global Definition /////////////////////////////////////////
#define SBLIMIT                 32
#define SSLIMIT                 18

/* MPEG Header Definitions - Mode Values */
#define MPG_MD_STEREO           0
#define MPG_MD_JOINT_STEREO     1
#define MPG_MD_DUAL_CHANNEL     2
#define MPG_MD_MONO             3

/* MPEG Header Definitions - ID Bit Values */
#define MPEG_AUDIO_ID           1
#define MPEG_PHASE2_LSF         0

// Global Type Definitions ////////////////////////////////////

typedef struct {
    unsigned ReadIndex;
    unsigned WriteIndex;
    unsigned char Buffer[FRAME_SIZE];
    } Mp3DataBuffer;

typedef struct {
    int Stereo;
    int SamplingFrequency;
    const char *FileName;
    } PcmFileInfo;

typedef struct {
    short Data[2][SSLIMIT][SBLIMIT];
    } PcmSampleFrame;

////////////////////////////////////////////////////////////////////////////////
//  Decode Header Information
//
//  this function read 20-Bits Header Information after SYNC and
//  places the bit stream pointer right after the header info.
//
//   | Frame Start (Byte Aligned)
//   |<-- SYNC ---->|
//   +--------------+-----------------------------------------+-----
//   |1111_1111_1111| Header Info                             |
//   +--------------+-----------------------------------------+-----
//   |<----------------- 32-Bits Header --------------------->|
//
//  Header (32-Bits)
//      +- SyncWord (12)
//      +- Version ID (1)
//      +- Layer (2) - Layer I("11")/II("10")/III("01")
//      +- Protection_Bit (1) - CRC Protection ('1' - NO, '0' - YES)
//      +- Bitrate_Index (4) - Bitrate Index ("1100" - 128k)
//      +- Sampling_Frequency (2) - "00"(22.05),"01"(24),"10"(16),"11"(Reserved)
//      |                           "00"(44.1), "01"(48),"10"(32),"11"(Reserved)
//      +- Padding_Bit (1)
//      +- Private_Bit (1)
//      +- Mode (2) - Stereo("00"),J-Stereo("01"),Dual Ch(10),Mono(11) Mode
//      +- Mode_Extension (2) - "00"-Non-Stereo,"01"-Intensity Stereo,
//      |                       "10"-MS-Stereo, "11"-both
//      |       when mode=J-Stereo, Intensity Stereo and MS-Stereo "off" both.
//      +- Copyright (1)
//      +- original/copy (1)
//      +- Emphasis (2)
//

typedef unsigned long Mp3FrameHeader;

// Raw frame header info (extracted directly from MP3 stream)
#define Mp3HeaderSync(      headerWord ) ( (headerWord >> 20) & 0xfff )
#define Mp3HeaderVersion(   headerWord ) ( (headerWord >> 19) & 0x1   )
#define Mp3HeaderLayer(     headerWord ) ( \
                                      4 - ((headerWord >> 17) & 0x3)  )
#define Mp3HeaderProtection(headerWord ) ( \
                                        ! ((headerWord >> 16) & 0x1)  )
#define Mp3HeaderBitrateID( headerWord ) ( (headerWord >> 12) & 0xf   )
#define Mp3HeaderSamplingFrequency( \
                            headerWord ) ( (headerWord >> 10) & 0x3   )
#define Mp3HeaderPadding(   headerWord ) ( (headerWord >>  9) & 0x1   )
#define Mp3HeaderPrivate(   headerWord ) ( (headerWord >>  8) & 0x1   )
#define Mp3HeaderMode(      headerWord ) ( (headerWord >>  6) & 0x3   )
#define Mp3HeaderModeExt(   headerWord ) ( (headerWord >>  4) & 0x3   )
#define Mp3HeaderCopyright( headerWord ) ( (headerWord >>  3) & 0x1   )
#define Mp3HeaderOriginal(  headerWord ) ( (headerWord >>  2) & 0x1   )
#define Mp3HeaderEmphasis(  headerWord ) ( (headerWord      ) & 0x3   )

// Auxiliary frame header info (derived from raw header)
#define Mp3HeaderStereo(    headerWord ) ( \
    Mp3HeaderMode(headerWord) == MPG_MD_MONO ? 0 : 1 )
#define Mp3HeaderJsbound(   headerWord ) ( \
    Mp3HeaderModeExt(headerWord) ? 1 << Mp3HeaderModeExt(headerWord+1) : 0 )

/* Layer III side information. */
typedef struct {
    unsigned Part2_3Length;
    unsigned BigValues;
    unsigned GlobalGain;
    unsigned ScalefacCompress;
    unsigned WindowSwitchingFlag;
    unsigned BlockType;
    unsigned MixedBlockFlag;
    unsigned TableSelect[3];
    unsigned SubblockGain[3];
    unsigned Region0Count;
    unsigned Region1Count;
    unsigned Preflag;
    unsigned ScalefacScale;
    unsigned Count1tableSelect;
    } Mp3SideHeader;

#define Mp3Scfsi( scfsiBits, id ) ( (scfsiBits>>(3-id)) & 0x1 )

typedef struct {
    unsigned MainDataBegin;
    unsigned PrivateBits;
    struct {
        unsigned ScfsiBits;
        Mp3SideHeader Gr[2];
        } Ch[2];
    } Mp3SideHeaderBlock;

typedef struct {
    Mp3FrameHeader     FrameHeader;
    Mp3SideHeaderBlock SideHeaderBlock;
    } Mp3Header;

/* Layer III scale factors. */
typedef struct {
    int l[23];            /* [cb] */
    int s[3][13];         /* [window][cb] */
    } III_scalefac_t[2];  /* [ch] */

#endif
