///////////////////////////////////////////////////////////////////////////
// Filename : Mp3Decoder.hxx
// Project  : Ans_MP3_Player/MP3_Decoder
// Author   : GoodKook / goodkook@nms.anslab.co.kr
// Company  : AnsLab Co. / http://www.anslab.co.kr
// Revision : 5-4-01: Modified for interface to untimed SystemC testbench
//                    - John Stickley, Ikos Systems, Inc. (johns@ikos.com)
// Comment  :
//
//

//===========================================================================
// Mp3Decoder                                                   johnS 4-21-01
//===========================================================================

#include <systemc.h>

#include "Mp3Common.h"
#include "FrameBuffer.hxx"

#include "sc_III_Decoder_ctl.h"

#include "sc_III_Decoder_Mem.h"
#include "sc_III_ScaleFactor.h"
#include "sc_III_Huffman.h"
#include "sc_III_Dequantizer.h"
#include "sc_III_Stereo.h"
#include "sc_III_Reorder.h"
#include "sc_III_Antialias.h"
#include "sc_III_Hybrid.h"
#include "sc_III_Subband.h"

SC_MODULE( ResetGenerator ){

    //-------------------------------------------------------
    // Timed port declarations
    sc_in_clk      Clock;
    sc_out<sc_bit> Reset;

    //-------------------------------------------------------
    // Context declarations
    private:
        int dResetCounter;

    //-------------------------------------------------------
    // Thread declarations
    private:
        void clockThread(){
            Reset = '0';
            while( dResetCounter ){
                if( Clock.read() == (sc_bit)'1' )
                    dResetCounter--;
                wait();
                }
            Reset = '1';
            }

    //-------------------------------------------------------
    // Constructor
    public:
        SC_CTOR( ResetGenerator ) : dResetCounter(RESET_CYCLES) {
            //--------------------------------------
            // Thread bindings

            SC_THREAD( clockThread );
            sensitive << Clock;
        }
    };

SC_MODULE( Mp3Decoder ){

    //--------------------------------------
    // Untimed port declarations.
    sc_slave<>  PlayIn, StopIn;                         // <-- slave
    sc_master<> DoneOut;                                // --> master

    sc_master<>                        RequestFrameOut; // --> master
    sc_inmaster<const Mp3Header *>     FrameHeaderIn;   // <-- inmaster
    sc_inmaster<const Mp3DataBuffer *> DataBufferIn;    // <-- inmaster

    sc_outmaster<const PcmSampleFrame *> PcmFrameOut;   // --> outmaster


    //--------------------------------------
    // Timed signal declarations.
    sc_signal<sc_bit> reset;

    sc_signal<sc_bit> fillStream, streamFull;
    sc_signal<sc_bit> memInit,    memInitDone;

    sc_signal<sc_bit>
        hdrReady,        hdrStart,        sclReady,      sclStart,
        huffReady,       huffStart,       deqReady,      deqStart,
        stereoReady,     stereoStart,     reorderReady,  reorderStart,
        antialiasReady,  antialiasStart,  hybridReady,   hybridStart,
        subbandReady,    subbandStart;

    sc_signal<sc_bit> channel, granule;

    sc_signal<Mp3FrameHeader>         frameHeader;
    sc_signal<const Mp3SideHeader *>  sideHeader;
    sc_signal<unsigned>               scfsiBits;

    sc_signal<sc_bit> preflagOut;
    sc_signal<unsigned long> part2Start;
 
    //-------------------------------------------------------
    // Thread declarations
    private:
        void playInThread(){ }
        void stopInThread(){ }

    //-------------------------------------------------------
    // Module declarations

    public:
    FrameBuffer     *frameBuffer;
    decoder_ctl     *u_decoder_ctl;
    decoder_mem     *u_decoder_mem;
    iii_scalefactor *u_scale;
    iii_huffman     *u_huffman;
    iii_dequantizer *u_dequantizer;
    iii_stereo      *u_stereo;
    iii_reorder     *u_reorder;
    iii_antialias   *u_antialias;
    iii_hybrid      *u_hybrid;
    iii_subband     *u_subband;
    ResetGenerator  *resetGenerator;

    SC_CTOR( Mp3Decoder ){

        frameBuffer = new FrameBuffer( "frameBuffer" );
            frameBuffer->RequestFrameOut(RequestFrameOut); // --> master
            frameBuffer->FrameHeaderIn(FrameHeaderIn);     // <-- inslave
            frameBuffer->DataBufferIn(DataBufferIn);       // <-- inslave

            frameBuffer->Clock( clock );                   // <--
            frameBuffer->Reset( reset );                   // <--

            frameBuffer->FillStream( fillStream );         // <--
            frameBuffer->StreamFull( streamFull );         // -->

            frameBuffer->HdrStart( hdrStart );             // <--
            frameBuffer->HdrReady( hdrReady );             // -->

            frameBuffer->Channel( channel );               // -->
            frameBuffer->Granule( granule );               // -->

            frameBuffer->FrameHeader( frameHeader );       // -->
            frameBuffer->SideHeader( sideHeader );         // -->
            frameBuffer->ScfsiBits( scfsiBits );           // -->

        GlobalFrameBuffer = frameBuffer;

        u_decoder_ctl = new decoder_ctl( "u_decoder_ctl" );
            u_decoder_ctl->PlayIn( PlayIn );                  // <-- slave
            u_decoder_ctl->StopIn( StopIn );                  // <-- slave
            u_decoder_ctl->DoneOut( DoneOut );                // --> master

            u_decoder_ctl->clk( clock );                      // <--
            u_decoder_ctl->rst( reset );                      // <--

            u_decoder_ctl->m_init_done( memInitDone );        // <--
            u_decoder_ctl->mem_init( memInit );               // -->

            u_decoder_ctl->stream_full( streamFull );         // <--
            u_decoder_ctl->fill_stream( fillStream );         // -->

            u_decoder_ctl->hdr_ready( hdrReady );             // <--
            u_decoder_ctl->hdr_start( hdrStart );             // -->
            u_decoder_ctl->scl_ready( sclReady );             // <--
            u_decoder_ctl->scl_start( sclStart );             // -->
            u_decoder_ctl->huff_ready( huffReady );           // <--
            u_decoder_ctl->huff_start( huffStart );           // -->
            u_decoder_ctl->deq_ready( deqReady );             // <--
            u_decoder_ctl->deq_start( deqStart );             // -->
            u_decoder_ctl->stereo_ready( stereoReady );       // <--
            u_decoder_ctl->stereo_start( stereoStart );       // -->
            u_decoder_ctl->reorder_ready( reorderReady );     // <--
            u_decoder_ctl->reorder_start( reorderStart );     // -->
            u_decoder_ctl->antialias_ready( antialiasReady ); // <--
            u_decoder_ctl->antialias_start( antialiasStart ); // -->
            u_decoder_ctl->hybrid_ready( hybridReady );       // <--
            u_decoder_ctl->hybrid_start( hybridStart );       // -->
            u_decoder_ctl->subband_ready( subbandReady );     // <--
            u_decoder_ctl->subband_start( subbandStart );     // -->

            u_decoder_ctl->channel(channel);                 // -->
            u_decoder_ctl->granule(granule);                 // -->

            u_decoder_ctl->FrameHeader(frameHeader);         // <--
 
        u_decoder_mem = new decoder_mem("u_decoder_mem");
            u_decoder_mem->clk(clock);                       // <--
            u_decoder_mem->rst(reset);                       // <--
            u_decoder_mem->mem_init(memInit);                // <--
            u_decoder_mem->m_init_done(memInitDone);         // -->
        GlobalDecoderMem = u_decoder_mem;
 
        u_scale =  new iii_scalefactor("u_scale");
            u_scale->clk(clock);                             // <--
            u_scale->rst(reset);                             // <--
            u_scale->start(sclStart);                        // <--
            u_scale->ready(sclReady);                        // -->

            u_scale->ch(channel);                            // <--
            u_scale->gr(granule);                            // <--

            u_scale->FrameHeader(frameHeader);               // <--
            u_scale->SideHeader(sideHeader);                 // <--
            u_scale->ScfsiBits(scfsiBits);                   // <--
            u_scale->preflag_out(preflagOut);                // -->
            u_scale->part2_start(part2Start);                // -->
   
        u_huffman = new iii_huffman("u_huffman");
            u_huffman->clk(clock);                           // <--
            u_huffman->rst(reset);                           // <--
            u_huffman->start(huffStart);                     // <--
            u_huffman->ready(huffReady);                     // -->
            u_huffman->part2_start(part2Start);              // <--

            u_huffman->ch(channel);                          // <--
            u_huffman->gr(granule);                          // <--

            u_huffman->FrameHeader(frameHeader);             // <--
            u_huffman->SideHeader(sideHeader);               // <--
 
        u_dequantizer = new iii_dequantizer("u_dequantizer");
            u_dequantizer->clk(clock);                       // <--
            u_dequantizer->rst(reset);                       // <--
            u_dequantizer->start(deqStart);                  // <--
            u_dequantizer->ready(deqReady);                  // -->

            u_dequantizer->preflag(preflagOut);              // <--
            u_dequantizer->channel(channel);                 // <--

            u_dequantizer->FrameHeader(frameHeader);         // <--
            u_dequantizer->SideHeader(sideHeader);           // <--
 
        u_stereo = new iii_stereo("u_stereo");
            u_stereo->clk(clock);                            // <--
            u_stereo->rst(reset);                            // <--
            u_stereo->start(stereoStart);                    // <--
            u_stereo->ready(stereoReady);                    // -->

            u_stereo->FrameHeader(frameHeader);              // <--
            u_stereo->SideHeader(sideHeader);                // <--
   
        u_reorder = new iii_reorder("u_reorder");
            u_reorder->clk(clock);                           // <--
            u_reorder->rst(reset);                           // <--
            u_reorder->start(reorderStart);                  // <--
            u_reorder->ready(reorderReady);                  // -->
            u_reorder->channel(channel);                     // <--

            u_reorder->FrameHeader(frameHeader);             // <--
            u_reorder->SideHeader(sideHeader);               // <--
   
        u_antialias = new iii_antialias("u_antialias");
            u_antialias->clk(clock);                         // <--
            u_antialias->rst(reset);                         // <--
            u_antialias->start(antialiasStart);              // <--
            u_antialias->ready(antialiasReady);              // -->

            u_antialias->channel(channel);                   // <--

            u_antialias->FrameHeader(frameHeader);           // <--
            u_antialias->SideHeader(sideHeader);             // <--
   
        u_hybrid = new iii_hybrid("u_hybrid");
            u_hybrid->clk(clock);                            // <--
            u_hybrid->rst(reset);                            // <--
            u_hybrid->start(hybridStart);                    // <--
            u_hybrid->ready(hybridReady);                    // -->

            u_hybrid->channel(channel);                      // <--

            u_hybrid->FrameHeader(frameHeader);              // <--
            u_hybrid->SideHeader(sideHeader);                // <--

        u_subband = new iii_subband("u_subband");
            u_subband->PcmFrameOut(PcmFrameOut);             // --> outmaster

            u_subband->clk(clock);                           // <--
            u_subband->rst(reset);                           // <--
            u_subband->start(subbandStart);                  // <--
            u_subband->ready(subbandReady);                  // -->

            u_subband->FrameHeader(frameHeader);             // -->


        resetGenerator = new ResetGenerator( "resetGenerator" );
            resetGenerator->Clock( clock );                  // <--
            resetGenerator->Reset( reset );                  // -->
 
        // Sensitize to dummy threads since SystemC v1.1 requires
        // sensitivity even on hierarchical pass throughs.
        SC_SLAVE( playInThread, PlayIn );
        SC_SLAVE( stopInThread, StopIn );
        }
    };
