#include <stdio.h>
#include <stdlib.h>

#include <fstream>

#include "WavImage.h"

//--------------------------------------------------------------------------
//  Exceptions.
//

    static void
errorFileOpen( const string &fileName, const char *forWhat ){
    string message( "Cannot open file " );
    message += fileName; message += " for "; message += forWhat;
        throw message; }

    static void
errorFileFormat( const string &fileName ){
    string message( "Invalid format while reading file " );
    message += fileName; message += ".";
        throw message; }

//--------------------------------------------------------------------------
//  WavFile Image
//

WavImage::~WavImage() {
    if( dIfs ){
        dIfs->close();
        delete dIfs;
        }
    if( dOfs ){
        // if( more samples were written than originally predicted )
        //     Recompute the file paraments.
        //     Lseek back and re-save the header.
        if( dActualNumSamples > dNumSamples ){
            Setup(
                dFmtChannels, dFmtBitsPerSample,
                dActualNumSamples, dFmtSamplesPerSec );
            Save( NULL );
            }
        dOfs->close();
        delete dOfs;
        }
    }

void WavImage::Setup(
    unsigned short channels, unsigned short bitsPerSample,
    unsigned long numSamples, unsigned long samplesPerSec )
{
    dFmtChunkLength = 16;
    dFmtTag = 1;
    dFmtChannels = channels;
    dFmtSamplesPerSec = samplesPerSec;
    dFmtBitsPerSample = bitsPerSample;
    dFmtBlockAlign = dFmtChannels * (dFmtBitsPerSample/8);
    dFmtBytesPerSec = samplesPerSec * dFmtBlockAlign;

    dNumSamples = numSamples;
    dDataLength = dNumSamples * (dFmtBitsPerSample/8) * dFmtChannels;
    dTimeSpan = dNumSamples / dFmtSamplesPerSec;
}

static const int PCM = 1;       //  Only PCM format supported.
static const int PCMChunkLength1 = 16;
static const int PCMChunkLength2 = 18;

void WavImage::Load( const char *fileName ) {
    string message;

    char RIFF[4]; RIFF[0] = 0;
    char WAVE[4]; WAVE[0] = 0;

    dFileName = fileName;

    if( dIfs ){
        dIfs->close();
        delete dIfs;
        }
    dIfs = new ifstream( dFileName.data(), ios::binary|ios::in );

    if( !dIfs->good() )
        errorFileOpen( dFileName, "reading" );

    dIfs->read( RIFF, sizeof(RIFF) );
    Read( dFileLength );
    dIfs->read( WAVE, sizeof(WAVE) );

    if( !dIfs->good() ) errorFileFormat( dFileName );

    if( strncmp( RIFF, "RIFF", sizeof(RIFF) )
            || strncmp( WAVE, "WAVE", sizeof(WAVE) ) )
        errorFileFormat( dFileName );

    // Load fmt chunk.
    char FMT[4]; FMT[0] = 0;

    dIfs->read( FMT, sizeof(FMT) );

    if( strncmp( FMT, "fmt ", sizeof(FMT) ) )
        errorFileFormat( dFileName );

    Read( dFmtChunkLength );
    Read( dFmtTag );
    Read( dFmtChannels );
    Read( dFmtSamplesPerSec );
    Read( dFmtBytesPerSec );
    Read( dFmtBlockAlign );
    Read( dFmtBitsPerSample );

    if( !dIfs->good() )
        errorFileFormat( dFileName );

    else if( dFmtTag != PCM ){
        message = "Invalid format tag - only PCM supported"; throw message; }

    else if( dFmtChunkLength != PCMChunkLength1
            && dFmtChunkLength != PCMChunkLength2 ){
        message = "Invalid chunk length detected for PCM format chunk.";
        throw message; }

    else if( dFmtChannels != 1 && dFmtChannels != 2 ){
        message = "Only 1 or 2 channels are allowed.";
        throw message; }

    else if( dFmtBitsPerSample != 8 && dFmtBitsPerSample != 16 ){
        message = "Only bits per sample of 8 or 16 are allowed.";
        throw message; }

    else if( dFmtBlockAlign != dFmtChannels * dFmtBitsPerSample / 8 ){
        char buf[256];
        sprintf( buf, "Block align (%d) is inconsistent number of channels (%d) and sample size (%d bits).",
            dFmtBlockAlign, dFmtChannels, dFmtBitsPerSample );
        message = buf;
        throw message; }

    else if( dFmtBytesPerSec != dFmtSamplesPerSec * dFmtBlockAlign ){
        char buf[256];
        sprintf( buf, "Average byte rate (%d) is inconsistent sampling rate (%d samples/sec) and block align (%d).", dFmtSamplesPerSec, dFmtBlockAlign );
        message = buf;
        throw message;
        }

    //  Load the data section.
    char DATA[4]; DATA[0] = 0;

    dIfs->read( DATA, sizeof(DATA) );
    Read( dDataLength );

    if( !dIfs->good() ) errorFileFormat( dFileName );

    else if( strncmp( DATA, "data", 4 ) ) errorFileFormat( dFileName );

    else if( dDataLength != dFileLength
            - sizeof(WAVE) - sizeof(FMT)
            - sizeof(dFmtChunkLength) - dFmtChunkLength
            - sizeof(DATA) - sizeof(dDataLength) ){
        char buf[256];
        sprintf( buf, "Data length is (%d bytes) is inconsistent with file length (%d bytes).", dDataLength, dFileLength );
        message = buf;
        throw message;
        }

    dNumSamples = dDataLength / ( (dFmtBitsPerSample/8) * dFmtChannels );
    dTimeSpan = dNumSamples / dFmtSamplesPerSec;
}

void WavImage::Save( const char *fileName ){
    string message;

    // For a re-save of the header, the file name will be NULL )
    if( fileName == NULL ){
        if( dOfs == NULL ){
            message = "Error re-saving file header.  Output file should have already been opened.";
            throw message; }
        else
            // Seek file back to beginning.
            dOfs->seekp(0);
        }
    else {
        dFileName = fileName;
        if( dOfs ){
            dOfs->close();
            delete dOfs;
            }
#ifdef WIN32
        dOfs = new ofstream( dFileName.data(), ios::binary|ios::out );
#else
        dOfs = new ofstream( dFileName.data() );
#endif
        }
    ofstream &ofs = *dOfs;

    if( !ofs.good() )
        errorFileOpen( dFileName, "writing" );

    dFileLength = 
        strlen( "WAVE" )                        // Length of Header.
      + strlen("fmt ") + sizeof(dFmtChunkLength) + dFmtChunkLength
            // Length of fmt chunk.
      + strlen("data") + sizeof(dDataLength) + dDataLength;
            // Length of data chunk.

    //  Write the RIFF header.
    ofs << "RIFF";
    Write( dFileLength );
    ofs << "WAVE";

    //  Write the fmt chunk.
    ofs << "fmt ";
    Write( dFmtChunkLength );
    Write( dFmtTag );
    Write( dFmtChannels );
    Write( dFmtSamplesPerSec );
    Write( dFmtBytesPerSec );
    Write( dFmtBlockAlign );
    Write( dFmtBitsPerSample );

    //  Write the data.
    ofs << "data";
    Write( dDataLength );

    dActualNumSamples = 0;
}

//
// These overloaded read and write routines always write big endian (Intel)
// to the .wav file or read big endian from the file regardless of local
// machine type.
//

void WavImage::Write( unsigned short data ) {
    register unsigned char bytes[sizeof data];
    bytes[0] = data;
    bytes[1] = data >> 8;
    dOfs->write( (const char *)bytes, sizeof(bytes) );
    dActualNumSamples++;
}

void WavImage::Write( unsigned long data ) {
    register unsigned char bytes[sizeof data];
    bytes[0] = data;
    bytes[1] = data >> 8;
    bytes[2] = data >> 16;
    bytes[3] = data >> 24;
    dOfs->write( (const char *)bytes, sizeof(bytes) );
    dActualNumSamples++;
}

void WavImage::Write( short data ) {
    register unsigned char bytes[sizeof data];
    bytes[0] = data;
    bytes[1] = data >> 8;
    dOfs->write( (const char *)bytes, sizeof(bytes) );
    dActualNumSamples++;
}

void WavImage::Write( long data ) {
    register unsigned char bytes[sizeof data];
    bytes[0] = data;
    bytes[1] = data >> 8;
    bytes[2] = data >> 16;
    bytes[3] = data >> 24;
    dOfs->write( (const char *)bytes, sizeof(bytes) );
    dActualNumSamples++;
}

void WavImage::Read( unsigned char &data ){
    *dIfs >> data;
}

void WavImage::Read( unsigned short &data ){
    register unsigned char bytes[sizeof(unsigned short)];
    dIfs->read( (char *)bytes, sizeof(bytes) );
    data  = (unsigned short)(bytes[1]) << 8;
    data += bytes[0];
}

void WavImage::Read( unsigned long &data ){
    register unsigned char bytes[sizeof(unsigned long)];
    dIfs->read( (char *)bytes, sizeof(bytes) );
    data  = (unsigned long)(bytes[3]) << 24;
    data += (unsigned long)(bytes[2]) << 16;
    data += (unsigned long)(bytes[1]) << 8;
    data += bytes[0];
}

void WavImage::Read( char &data ){
    *dIfs >> data;
}

void WavImage::Read( short &data ){
    register unsigned char bytes[sizeof(short)];
    dIfs->read( (char *)bytes, sizeof(bytes) );
    data  = (unsigned short)(bytes[1]) << 8;
    data += bytes[0];
}

void WavImage::Read( long &data ){
    register unsigned char bytes[sizeof(long)];
    dIfs->read( (char *)bytes, sizeof(bytes) );
    data  = (unsigned long)(bytes[3]) << 24;
    data += (unsigned long)(bytes[2]) << 16;
    data += (unsigned long)(bytes[1]) << 8;
    data += bytes[0];
}
